import multiprocessing as mp
import sys

# try:
#     mp.set_start_method("spawn")
# except:
#     pass


def main(queue: mp.Queue,
         ncols: int = 4,
         minimum_height: int = 400,
         colors: list[tuple[int, int, int]] = []):
    from .mainwindow import MainWindow
    from .qt_compat import QtWidgets

    app = QtWidgets.QApplication(sys.argv)
    main = MainWindow(queue, ncols, minimum_height, colors)
    sys.exit(app.exec())


class Monitor():

    def __init__(self,
                 number_of_columns: int = 4,
                 minimum_height: int = 400,
                 colors: list[tuple[int, int, int]] = []):
        self.colors = [tuple(color) for color in colors]
        self.number_of_columns = number_of_columns
        self.minimum_height = minimum_height
        self.queue = mp.Queue(20)
        self.process = None
        self.start()

    def start(self):
        if self.process is not None and self.process.is_alive():
            return
        self.queue = mp.Queue(20)
        self.process = mp.Process(target=main,
                                  args=(self.queue, self.number_of_columns,
                                        self.minimum_height, self.colors))
        self.process.start()

    def _put(self, w: tuple):
        self.queue.put(w)

    def roll(self):
        self._put(("ROLL", None))

    def set_column_names(self, *arg):
        self._put(('PN', list(arg)))

    def add_point(self, *arg):
        self._put(('PD', list(arg)))

    def set_plots(self, arg):
        """
        arg: str, like "(x,y)" or "(x1,y1);(x2,y2);"
        """
        self._put(('PXY', str(arg)))

    def set_trace_column_names(self, *arg):
        self._put(('TN', list(arg)))

    def add_trace(self, *arg):
        self._put(('TD', list(arg)))

    def set_trace_plots(self, arg):
        """
        arg: str, like "(x,y)" or "(x1,y1);(x2,y2);"
        """
        self._put(('TXY', str(arg)))

    def __del__(self):
        try:
            self.process.kill()
        except:
            pass

    def is_alive(self):
        return self.process.is_alive()


_monitor = None


def get_monitor(auto_open=True):
    global _monitor

    if auto_open and (_monitor is None or not _monitor.is_alive()):
        _monitor = Monitor()

    return _monitor



import ast
import asyncio
import inspect
import warnings
from typing import Any, Callable

import dill

from .expression import Env, Expression


class Unpicklable:

    def __init__(self, obj):
        self.type = str(type(obj))
        self.id = id(obj)

    def __repr__(self):
        return f'<Unpicklable: {self.type} at 0x{id(self):x}>'


class TooLarge:

    def __init__(self, obj):
        self.type = str(type(obj))
        self.id = id(obj)

    def __repr__(self):
        return f'<TooLarge: {self.type} at 0x{id(self):x}>'


def dump_globals(ns=None, *, size_limit=10 * 1024 * 1024, warn=False):
    import __main__

    if ns is None:
        ns = __main__.__dict__

    namespace = {}

    for name, value in ns.items():
        try:
            buf = dill.dumps(value)
        except:
            namespace[name] = Unpicklable(value)
            if warn:
                warnings.warn(f'Unpicklable: {name} {type(value)}')
        if len(buf) > size_limit:
            namespace[name] = TooLarge(value)
            if warn:
                warnings.warn(f'TooLarge: {name} {type(value)}')
        else:
            namespace[name] = buf

    return namespace


def is_valid_identifier(s: str) -> bool:
    """
    Check if a string is a valid identifier.
    """
    try:
        ast.parse(f"f({s}=0)")
        return True
    except SyntaxError:
        return False


async def async_next(aiter):
    try:
        if hasattr(aiter, '__anext__'):
            return await aiter.__anext__()
        else:
            return next(aiter)
    except StopIteration:
        raise StopAsyncIteration from None


async def async_zip(*aiters):
    aiters = [
        ait.__aiter__() if hasattr(ait, '__aiter__') else iter(ait)
        for ait in aiters
    ]
    try:
        while True:
            # 使用 asyncio.gather 等待所有异步生成器返回下一个元素
            result = await asyncio.gather(*(async_next(ait) for ait in aiters))
            yield tuple(result)
    except StopAsyncIteration:
        # 当任一异步生成器耗尽时停止迭代
        return


async def call_function(func: Callable | Expression, variables: dict[str,
                                                                     Any]):
    if isinstance(func, Expression):
        env = Env()
        for name in func.symbols():
            if name in variables:
                if inspect.isawaitable(variables[name]):
                    variables[name] = await variables[name]
                env.variables[name] = variables[name]
            else:
                raise ValueError(f'{name} is not provided.')
        return func.eval(env)

    try:
        sig = inspect.signature(func)
    except:
        return func()
    args = []
    for name, param in sig.parameters.items():
        if param.kind == param.POSITIONAL_OR_KEYWORD:
            if name in variables:
                if inspect.isawaitable(variables[name]):
                    variables[name] = await variables[name]
                args.append(variables[name])
            elif param.default is not param.empty:
                args.append(param.default)
            else:
                raise ValueError(f'parameter {name} is not provided.')
        elif param.kind == param.VAR_POSITIONAL:
            raise ValueError('not support VAR_POSITIONAL')
        elif param.kind == param.VAR_KEYWORD:
            ret = func(**variables)
            if inspect.isawaitable(ret):
                ret = await ret
            return ret
    ret = func(*args)
    if inspect.isawaitable(ret):
        ret = await ret
    return ret



import dill
import numpy as np
import redis


def try_dumps(value):
    try:
        value_b = dill.dumps(value)
    except:
        value_b = dill.dumps(str(value))
    finally:
        return value_b


class redisClient(object):

    def __init__(self,
                 name,
                 server=None,
                 addr='redis://localhost:6379/0',
                 expire_time=172800):
        self._r = redis.Redis.from_url(addr) if server is None else server
        self.name = name
        self.expire_time = int(expire_time)  # 默认数据两天过期，单位秒

    def delete(self):
        self._r.delete(self.name)


class redisString(redisClient):
    '''Redis String client'''

    def set(self, value):
        if value is not None:
            value_b = try_dumps(value)
            self._r.set(self.name, value_b, ex=self.expire_time)

    def get(self):
        v_b = self._r.get(self.name)
        value = dill.loads(v_b) if v_b is not None else None
        return value

    @property
    def data(self):
        return self.get()


class redisList(redisClient):
    '''Redis List client'''

    def add(self, *arg):
        arg_b = [dill.dumps(i) for i in arg]
        self._r.rpush(self.name, *arg_b)
        self._r.expire(self.name, self.expire_time)

    def read(self, start=0, end=-1):
        data_b = self._r.lrange(self.name, start, end)
        data = [dill.loads(i) for i in data_b]
        return data

    @property
    def size(self):
        return self._r.llen(self.name)

    @property
    def data(self):
        return self.read()


class redisSet(redisClient):
    '''Redis Set client'''

    def add(self, *arg):
        arg_b = {dill.dumps(i) for i in arg}
        self._r.sadd(self.name, *arg_b)
        self._r.expire(self.name, self.expire_time)

    def read(self):
        data_b = self._r.smembers(self.name)
        data = {dill.loads(i) for i in data_b}
        return data

    @property
    def size(self):
        return self._r.scard(self.name)

    @property
    def data(self):
        return self.read()


class redisZSet(redisClient):
    '''有序集合'''

    def __init__(self,
                 name,
                 server=None,
                 addr='redis://localhost:6379/0',
                 expire_time=172800):
        super().__init__(name, server, addr, expire_time)
        self.__score = 0

    def delete(self):
        super().delete()
        self.__score = 0

    def add(self, *elements):
        mapping = {}
        for ele in elements:
            ele_b = dill.dumps(ele)
            self.__score += 1
            mapping.update({ele_b: self.__score})
        self._r.zadd(self.name, mapping, nx=True)  # 只添加新元素
        self._r.expire(self.name, self.expire_time)

    def read(self, start=0, end=-1):
        data_b = self._r.zrange(self.name, start, end, withscores=False)
        data = [dill.loads(i) for i in data_b]
        return data

    @property
    def size(self):
        return self._r.zcard(self.name)

    @property
    def data(self):
        return self.read()


class redisHash(redisClient):
    '''Redis Hash client'''

    def add(self, **kw):
        kw_b = {k: try_dumps(v) for k, v in kw.items()}
        self._r.hmset(self.name, kw_b)
        self._r.expire(self.name, self.expire_time)

    def read(self):
        data_b = self._r.hgetall(self.name)
        data = {k_b.decode(): dill.loads(v_b) for k_b, v_b in data_b.items()}
        return data

    def get(self, key):
        '''读取Hash中的一个key'''
        v_b = self._r.hget(self.name, key)
        value = dill.loads(v_b) if v_b is not None else None
        return value

    @property
    def size(self):
        return self._r.hlen(self.name)

    @property
    def data(self):
        return self.read()


class redisArray(redisClient):
    '''Redis np.array client'''

    def __init__(self,
                 name,
                 server=None,
                 addr='redis://localhost:6379/0',
                 expire_time=172800,
                 dtype='complex128'):
        super().__init__(name, server, addr, expire_time)
        _r_dtype = self._r.get(f'{name}.dtype')
        if _r_dtype is None:
            self._r.set(f'{name}.dtype', dtype, ex=self.expire_time)
            self.dtype = dtype
        else:
            self.dtype = _r_dtype

    def delete(self):
        self._r.delete(self.name)
        self._r.delete(f'{self.name}.dtype')

    def add(self, *args):
        for arg in args:
            buf = np.asarray(arg).astype(self.dtype).tobytes()
            # self._r.append(self.name, buf)
            self._r.rpush(self.name, buf)
        self._r.expire(self.name, self.expire_time)

    def read(self):
        # buf = self._r.get(self.name)
        buf_list = self._r.lrange(self.name, 0, -1)
        buf = b''.join(buf_list)
        data = np.frombuffer(buf,
                             dtype=self.dtype) if buf is not None else None
        return data

    @property
    def size(self):
        array = self.data
        if array is None:
            return 0
        else:
            return array.size

    @property
    def data(self):
        return self.read()



import hashlib
import pickle
from datetime import datetime, timezone
from functools import singledispatchmethod

from sqlalchemy import (Column, DateTime, Float, ForeignKey, Integer,
                        LargeBinary, String, Table, Text, create_engine)
from sqlalchemy.orm import (backref, declarative_base, relationship,
                            sessionmaker)
from sqlalchemy.orm.session import Session
from waveforms.security import InvalidKey, encryptPassword, verifyPassword


def utcnow():
    return datetime.now(timezone.utc)


Base = declarative_base()

# association table
user_roles = Table('user_roles', Base.metadata,
                   Column('user_id', ForeignKey('users.id'), primary_key=True),
                   Column('role_id', ForeignKey('roles.id'), primary_key=True))

record_reports = Table(
    'record_reports', Base.metadata,
    Column('record_id', ForeignKey('records.id'), primary_key=True),
    Column('report_id', ForeignKey('reports.id'), primary_key=True))

comment_tags = Table(
    'comment_tags', Base.metadata,
    Column('comment_id', ForeignKey('comments.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

snapshot_tags = Table(
    'snapshot_tags', Base.metadata,
    Column('snapshot_id', ForeignKey('snapshots.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

record_tags = Table(
    'record_tags', Base.metadata,
    Column('record_id', ForeignKey('records.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

report_tags = Table(
    'report_tags', Base.metadata,
    Column('report_id', ForeignKey('reports.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

sample_tags = Table(
    'sample_tags', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

sample_reports = Table(
    'sample_reports', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('report_id', ForeignKey('reports.id'), primary_key=True))

sample_records = Table(
    'sample_records', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('record_id', ForeignKey('records.id'), primary_key=True))

sample_comments = Table(
    'sample_comments', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))

sample_transfer_comments = Table(
    'sample_transfer_comments', Base.metadata,
    Column('transfer_id', ForeignKey('sample_transfer.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))

report_comments = Table(
    'report_comments', Base.metadata,
    Column('report_id', ForeignKey('reports.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))

record_comments = Table(
    'record_comments', Base.metadata,
    Column('record_id', ForeignKey('records.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))


class Role(Base):
    __tablename__ = 'roles'

    id = Column(Integer, primary_key=True)
    name = Column(String)

    users = relationship('User', secondary=user_roles, back_populates='roles')

    def __repr__(self):
        return f"Role(name='{self.name}')"


class User(Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    hashed_password = Column(LargeBinary(64))
    fullname = Column(String)

    roles = relationship('Role', secondary=user_roles, back_populates='users')
    attachments = relationship('Attachment', back_populates='user')
    comments = relationship('Comment', back_populates='user')

    def setPassword(self, password):
        self.hashed_password = encryptPassword(password)

    def verify(self, password):
        try:
            verifyPassword(password, self.hashed_password)
            return True
        except InvalidKey:
            return False

    def __repr__(self):
        return f"User(name='{self.name}')"


class ParameterEdge(Base):
    __tablename__ = 'parameter_edges'
    parent_id = Column(ForeignKey('parameters.id'), primary_key=True)
    child_id = Column(ForeignKey('parameters.id'), primary_key=True)
    key = Column(String(50), primary_key=True)


class Parameter(Base):
    __tablename__ = 'parameters'

    id = Column(Integer, primary_key=True)
    type = Column(String)
    unit = Column(String)
    integer = Column(Integer)
    real = Column(Float)
    imag = Column(Float)
    string = Column(String)
    buff = Column(LargeBinary)

    children = relationship("ParameterEdge",
                            foreign_keys=[ParameterEdge.parent_id],
                            backref="parent")
    parents = relationship("ParameterEdge",
                           foreign_keys=[ParameterEdge.child_id],
                           backref="child")

    @property
    def value(self):
        if self.type == 'integer':
            return self.integer
        elif self.type == 'real':
            return self.real
        elif self.type == 'complex':
            return self.real + 1j * self.imag
        elif self.type == 'string':
            return self.string
        elif self.type == 'buffer':
            return self.buff
        elif self.type == 'boolean':
            return self.integer == 1
        elif self.type is None or self.type == 'none':
            return None
        elif self.type == 'tree':
            return self.export()
        else:
            return pickle.loads(self.buff)

    @value.setter
    def value(self, value):
        self._set_value(value)

    @singledispatchmethod
    def _set_value(self, value):
        self.buff = pickle.dumps(value)
        self.type = 'pickle'

    @_set_value.register
    def _(self, value: int):
        self.integer = value
        self.type = 'integer'

    @_set_value.register
    def _(self, value: bool):
        self.integer = int(value)
        self.type = 'boolean'

    @_set_value.register
    def _(self, value: float):
        self.real = value
        self.type = 'real'

    @_set_value.register
    def _(self, value: complex):
        self.real = value.real
        self.imag = value.imag
        self.type = 'complex'

    @_set_value.register
    def _(self, value: str):
        self.string = value
        self.type = 'string'

    @_set_value.register
    def _(self, value: bytes):
        self.buff = value
        self.type = 'buffer'

    @_set_value.register
    def _(self, value: None):
        self.type = 'none'

    def export(self):
        if self.type == 'tree':
            return {a.key: a.child.export() for a in self.children}
        else:
            return self.value


class Snapshot(Base):
    __tablename__ = 'snapshots'

    id = Column(Integer, primary_key=True)
    root_id = Column(ForeignKey('parameters.id'))
    previous_id = Column(Integer, ForeignKey('snapshots.id'))

    followers = relationship("Snapshot",
                             backref=backref('previous', remote_side=[id]))

    root = relationship('Parameter')
    tags = relationship('Tag',
                        secondary=snapshot_tags,
                        back_populates='snapshots')

    def export(self):
        return self.root.export()


class ReportParameters(Base):
    __tablename__ = 'report_parameters'
    parent_id = Column(ForeignKey('reports.id'), primary_key=True)
    child_id = Column(ForeignKey('parameters.id'), primary_key=True)
    key = Column(String(50), primary_key=True)

    parameter = relationship('Parameter')

    @property
    def value(self):
        return self.parameter.value


class Tag(Base):
    __tablename__ = 'tags'

    id = Column(Integer, primary_key=True)
    text = Column(String, unique=True)

    comments = relationship('Comment',
                            secondary=comment_tags,
                            back_populates='tags')
    records = relationship('Record',
                           secondary=record_tags,
                           back_populates='tags')
    reports = relationship('Report',
                           secondary=report_tags,
                           back_populates='tags')
    samples = relationship('Sample',
                           secondary=sample_tags,
                           back_populates='tags')
    snapshots = relationship('Snapshot',
                             secondary=snapshot_tags,
                             back_populates='tags')

    def __init__(self, text) -> None:
        super().__init__()
        self.text = text

    def __repr__(self):
        return f"Tag('{self.text}')"


class Comment(Base):
    __tablename__ = 'comments'

    id = Column(Integer, primary_key=True)
    text = Column(String)
    user_id = Column(Integer, ForeignKey('users.id'))
    ctime = Column(DateTime, default=utcnow)
    mtime = Column(DateTime, default=utcnow)
    atime = Column(DateTime, default=utcnow)
    parent_id = Column(Integer, ForeignKey('comments.id'))

    replies = relationship("Comment", lazy="joined", join_depth=2)
    user = relationship('User', back_populates='comments')
    tags = relationship('Tag',
                        secondary=comment_tags,
                        back_populates='comments')
    attachments = relationship('Attachment', back_populates='comment')


class Attachment(Base):
    __tablename__ = 'attachments'

    id = Column(Integer, primary_key=True)
    filename = Column(String)
    mime_type = Column(String, default='application/octet-stream')
    user_id = Column(Integer, ForeignKey('users.id'))
    comment_id = Column(Integer, ForeignKey('comments.id'))
    ctime = Column(DateTime, default=utcnow)
    mtime = Column(DateTime, default=utcnow)
    atime = Column(DateTime, default=utcnow)
    size = Column(Integer)
    sha1 = Column(String)
    description = Column(Text)

    user = relationship('User', back_populates='attachments')
    comment = relationship('Comment', back_populates='attachments')


class InputText(Base):
    __tablename__ = 'inputs'

    id = Column(Integer, primary_key=True)
    hash = Column(LargeBinary(20), index=True)
    text_field = Column(Text, unique=True)

    @property
    def text(self):
        return self.text_field

    @text.setter
    def text(self, text):
        self.hash = hashlib.sha1(text.encode('utf-8')).digest()
        self.text_field = text

    def __repr__(self) -> str:
        return self.text


class Cell(Base):
    __tablename__ = 'cells'

    id = Column(Integer, primary_key=True)
    notebook_id = Column(Integer, ForeignKey("notebooks.id"))
    index = Column(Integer)
    ctime = Column(DateTime, default=utcnow)
    ftime = Column(DateTime, default=utcnow)
    input_id = Column(Integer, ForeignKey("inputs.id"))

    notebook = relationship("Notebook", back_populates="cells")
    input = relationship("InputText")

    def __repr__(self) -> str:
        return str(f"Cell(index={self.index}, input='{self.input}')")


class Notebook(Base):
    __tablename__ = 'notebooks'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    ctime = Column(DateTime, default=utcnow)
    atime = Column(DateTime, default=utcnow)

    cells = relationship("Cell",
                         order_by=Cell.index,
                         back_populates="notebook")


class Sample(Base):
    __tablename__ = 'samples'

    id = Column(Integer, primary_key=True)
    name = Column(String)

    account_id = Column(Integer, ForeignKey("sample_accounts.id"))

    tags = relationship("Tag", secondary=sample_tags, back_populates="samples")
    records = relationship("Record",
                           secondary=sample_records,
                           back_populates="samples")
    reports = relationship("Report",
                           secondary=sample_reports,
                           back_populates="samples")
    transfers = relationship("SampleTransfer",
                             order_by="SampleTransfer.ctime",
                             back_populates="sample")
    account = relationship("SampleAccount", back_populates="samples")
    comments = relationship("Comment", secondary=sample_comments)


class SampleAccountType(Base):
    __tablename__ = 'sample_account_types'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)

    accounts = relationship("SampleAccount", back_populates="type")


class SampleAccount(Base):
    __tablename__ = 'sample_accounts'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    type_id = Column(Integer, ForeignKey("sample_account_types.id"))
    description = Column(String)

    type = relationship("SampleAccountType", back_populates="accounts")

    samples = relationship("Sample", back_populates="account")


class SampleTransfer(Base):
    __tablename__ = 'sample_transfer'

    id = Column(Integer, primary_key=True)
    sample_id = Column(Integer, ForeignKey("samples.id"))
    user_id = Column(Integer, ForeignKey("users.id"))
    ctime = Column(DateTime, default=utcnow)
    debtor_id = Column(Integer, ForeignKey("sample_accounts.id"))
    creditor_id = Column(Integer, ForeignKey("sample_accounts.id"))

    user = relationship("User")
    sample = relationship("Sample", back_populates="transfers")
    debtor = relationship("SampleAccount", foreign_keys=[debtor_id])
    creditor = relationship("SampleAccount", foreign_keys=[creditor_id])
    comments = relationship("Comment", secondary=sample_transfer_comments)


class Config(Base):
    __tablename__ = 'configs'

    id = Column(Integer, primary_key=True)
    hash = Column(LargeBinary(20), index=True)
    file = Column(String)
    content_type = Column(String, default='application/pickle')
    ctime = Column(DateTime, default=utcnow)
    atime = Column(DateTime, default=utcnow)

    records = relationship("Record", back_populates="config")

    def __init__(self, data: bytes) -> None:
        self.hash = hashlib.sha1(data).digest()


class Record(Base):
    __tablename__ = 'records'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=utcnow)
    mtime = Column(DateTime, default=utcnow)
    atime = Column(DateTime, default=utcnow)
    user_id = Column(Integer, ForeignKey('users.id'))
    config_id = Column(Integer, ForeignKey('configs.id'))
    parent_id = Column(Integer, ForeignKey('records.id'))
    cell_id = Column(Integer, ForeignKey('cells.id'))

    app = Column(String)
    file = Column(String)
    content_type = Column(String, default='application/pickle')
    key = Column(String)

    parent = relationship("Record",
                          remote_side=[id],
                          back_populates="children")
    children = relationship("Record",
                            remote_side=[parent_id],
                            back_populates="parent")

    config = relationship("Config", back_populates="records")
    user = relationship("User")
    samples = relationship("Sample",
                           secondary=sample_records,
                           back_populates="records")
    cell = relationship("Cell")

    reports = relationship('Report',
                           secondary=record_reports,
                           back_populates='records')
    tags = relationship('Tag', secondary=record_tags, back_populates='records')
    comments = relationship('Comment', secondary=record_comments)


class Report(Base):
    __tablename__ = 'reports'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=utcnow)
    mtime = Column(DateTime, default=utcnow)
    atime = Column(DateTime, default=utcnow)
    user_id = Column(Integer, ForeignKey('users.id'))

    category = Column(String)
    title = Column(String)
    content = Column(Text)
    file = Column(String)
    key = Column(String)
    task_hash = Column(LargeBinary(32))

    user = relationship("User")
    samples = relationship("Sample",
                           secondary=sample_reports,
                           back_populates="reports")

    records = relationship('Record',
                           secondary=record_reports,
                           back_populates='reports')

    tags = relationship('Tag', secondary=report_tags, back_populates='reports')
    comments = relationship('Comment', secondary=report_comments)
    _parameters = relationship('ReportParameters')


def create_tables(engine, tables_only=False):
    Base.metadata.create_all(engine)
    if tables_only:
        return

    sys_role = Role(name='sys')
    kernel = User(name='BIG BROTHER')
    kernel.roles.append(sys_role)

    root_role = Role(name='root')
    admin_role = Role(name='admin')
    root_user = User(name='root')
    root_user.setPassword('123')
    root_user.roles.append(root_role)
    root_user.roles.append(admin_role)

    guest_role = Role(name='guest')
    guest_user = User(name='guest')
    guest_user.setPassword('')
    guest_user.roles.append(guest_role)

    t1 = SampleAccountType(name='factory')
    t2 = SampleAccountType(name='destroyed')
    t3 = SampleAccountType(name='storage')
    t4 = SampleAccountType(name='fridge')
    a = SampleAccount(name='destroyed')
    a.type = t2

    Session = sessionmaker(bind=engine)
    session = Session()

    session.add(root_user)
    session.add(guest_user)
    session.add(kernel)
    session.add_all([t1, t2, t3, t4, a])
    try:
        session.commit()
    except:
        session.rollback()



import itertools
from typing import Type

import numpy as np
import skopt
from skopt.space import Categorical, Integer, Real


class Space():

    def __init__(self, function, *args, **kwds):
        self.function = function
        self.args = args
        self.kwds = kwds

    def __repr__(self):
        if self.function == 'asarray':
            return repr(self.args[0])
        args = ', '.join(map(repr, self.args))
        kwds = ', '.join(f'{k}={v!r}' for k, v in self.kwds.items())
        return f"{self.function}({args}, {kwds})"

    def __len__(self):
        return len(self.toarray())

    @classmethod
    def fromarray(cls, space):
        if isinstance(space, Space):
            return space
        if isinstance(space, (list, tuple)):
            array = np.array(space)
        elif isinstance(space, np.ndarray):
            array = space
        try:
            a = np.linspace(array[0], array[-1], len(array), dtype=array.dtype)
            if np.allclose(a, array):
                return cls('linspace',
                           array[0],
                           array[-1],
                           len(array),
                           dtype=array.dtype)
        except:
            pass
        try:
            a = np.logspace(np.log10(array[0]),
                            np.log10(array[-1]),
                            len(array),
                            base=10,
                            dtype=array.dtype)
            if np.allclose(a, array):
                return cls('logspace',
                           np.log10(array[0]),
                           np.log10(array[-1]),
                           len(array),
                           base=10,
                           dtype=array.dtype)
        except:
            pass
        try:
            a = np.logspace(np.log2(array[0]),
                            np.log2(array[-1]),
                            len(array),
                            base=2,
                            dtype=array.dtype)
            if np.allclose(a, array):
                return cls('logspace',
                           np.log2(array[0]),
                           np.log2(array[-1]),
                           len(array),
                           base=2,
                           dtype=array.dtype)
        except:
            pass
        try:
            a = np.geomspace(array[0],
                             array[-1],
                             len(array),
                             dtype=array.dtype)
            if np.allclose(a, array):
                return cls('geomspace',
                           array[0],
                           array[-1],
                           len(array),
                           dtype=array.dtype)
        except:
            pass
        return space

    def toarray(self):
        return getattr(np, self.function)(*self.args, **self.kwds)


def logspace(start, stop, num=50, endpoint=True, base=10):
    return Space('logspace', start, stop, num, endpoint=endpoint, base=base)


def linspace(start, stop, num=50, endpoint=True):
    return Space('linspace', start, stop, num, endpoint=endpoint)


def geomspace(start, stop, num=50, endpoint=True):
    return Space('geomspace', start, stop, num, endpoint=endpoint)


class OptimizeSpace():

    def __init__(self, optimizer: 'Optimizer', space, suggestion=None):
        self.optimizer = optimizer
        self.space = space
        self.name = None
        if suggestion is not None and not isinstance(
                suggestion, (list, tuple, np.ndarray)):
            suggestion = [suggestion]
        self.suggestion = suggestion

    def __len__(self):
        return self.optimizer.maxiter


class Optimizer():

    def __init__(self,
                 scanner,
                 name: str,
                 level: int,
                 method: str | Type = skopt.Optimizer,
                 maxiter: int = 1000,
                 minimize: bool = True,
                 **kwds):
        self.scanner = scanner
        self.method = method
        self.maxiter = maxiter
        self.dimensions = {}
        self.name = name
        self.level = level
        self.kwds = kwds
        self.minimize = minimize
        self.suggestion = {}

    def create(self):
        opt = self.method(list(self.dimensions.values()), **self.kwds)

        def rvs(space):
            while True:
                yield space.rvs()[0]

        if self.suggestion:
            for suggestion in zip(*[
                    self.suggestion.get(key, rvs(space))
                    for key, space in self.dimensions.items()
            ]):
                opt.suggest(*suggestion)
        return opt

    def Categorical(self,
                    categories,
                    prior=None,
                    transform=None,
                    name=None,
                    suggestion=None) -> OptimizeSpace:
        return OptimizeSpace(self,
                             Categorical(categories, prior, transform, name),
                             suggestion)

    def Integer(self,
                low,
                high,
                prior="uniform",
                base=10,
                transform="normalize",
                name=None,
                dtype=np.int64,
                suggestion=None) -> OptimizeSpace:
        return OptimizeSpace(
            self, Integer(low, high, prior, base, transform, name, dtype),
            suggestion)

    def Real(self,
             low,
             high,
             prior="uniform",
             base=10,
             transform="normalize",
             name=None,
             dtype=float,
             suggestion=None) -> OptimizeSpace:
        return OptimizeSpace(
            self, Real(low, high, prior, base, transform, name, dtype),
            suggestion)

    def __getstate__(self) -> dict:
        state = self.__dict__.copy()
        del state['scanner']
        return state

    def __setstate__(self, state: dict) -> None:
        self.__dict__.update(state)
        self.scanner = None



import itertools
import lzma
import pickle
import sys
import uuid
import zipfile
from pathlib import Path
from threading import Lock
from types import EllipsisType

import dill
import numpy as np
import zmq

from qulab.sys.rpc.zmq_socket import ZMQContextManager

from .curd import get_config
from .space import OptimizeSpace, Space

_not_given = object()


def random_path(base):
    while True:
        s = uuid.uuid4().hex
        path = base / s[:2] / s[2:4] / s[4:6] / s[6:]
        if not path.exists():
            return path


def index_in_slice(slice_obj: slice | int, index: int):
    if isinstance(slice_obj, int):
        return slice_obj == index
    start, stop, step = slice_obj.start, slice_obj.stop, slice_obj.step
    if start is None:
        start = 0
    if step is None:
        step = 1
    if stop is None:
        stop = sys.maxsize

    if step > 0:
        return start <= index < stop and (index - start) % step == 0
    else:
        return stop < index <= start and (index - start) % step == 0


class BufferList():

    def __init__(self, file=None, slice=None):
        self._list = []
        self.lu = ()
        self.rd = ()
        self.inner_shape = ()
        self.file = file
        self._slice = slice
        self._lock = Lock()
        self._data_id = None
        self._sock = None

    def __repr__(self):
        return f"<BufferList: shape={self.shape}, lu={self.lu}, rd={self.rd}, slice={self._slice}>"

    def __getstate__(self):
        self.flush()
        if isinstance(self.file, Path):
            file = '/'.join(self.file.parts[-4:])
        else:
            file = self.file
        return {
            'file': file,
            'lu': self.lu,
            'rd': self.rd,
            'inner_shape': self.inner_shape,
        }

    def __setstate__(self, state):
        self.file = state['file']
        self.lu = state['lu']
        self.rd = state['rd']
        self.inner_shape = state['inner_shape']
        self._list = []
        self._slice = None
        self._lock = Lock()
        self._data_id = None
        self._sock = None

    @property
    def shape(self):
        return tuple([i - j
                      for i, j in zip(self.rd, self.lu)]) + self.inner_shape

    def flush(self):
        if not self._list:
            return
        if isinstance(self.file, Path):
            with self._lock:
                with open(self.file, 'ab') as f:
                    for item in self._list:
                        dill.dump(item, f)
                self._list.clear()

    def delete(self):
        if isinstance(self.file, Path):
            self.file.unlink()
            self.file = None

    def append(self, pos, value, dims=None):
        if dims is not None:
            if any([p != 0 for i, p in enumerate(pos) if i not in dims]):
                return
            pos = tuple([pos[i] for i in dims])
        self.lu = tuple([min(i, j) for i, j in zip(pos, self.lu)])
        self.rd = tuple([max(i + 1, j) for i, j in zip(pos, self.rd)])
        if hasattr(value, 'shape'):
            if self.inner_shape is None:
                self.inner_shape = value.shape
            elif self.inner_shape != value.shape:
                self.inner_shape = ()
        with self._lock:
            self._list.append((pos, value))
        if len(self._list) > 1000:
            self.flush()

    def _iter_file(self):
        if isinstance(self.file, Path) and self.file.exists():
            with self._lock:
                with open(self.file, 'rb') as f:
                    while True:
                        try:
                            pos, value = dill.load(f)
                            yield pos, value
                        except EOFError:
                            break
        elif isinstance(
                self.file, tuple) and len(self.file) == 2 and isinstance(
                    self.file[0], str) and self.file[0].endswith('.zip'):
            f, name = self.file
            with zipfile.ZipFile(f, 'r') as z:
                with z.open(name, 'r') as f:
                    while True:
                        try:
                            pos, value = dill.load(f)
                            yield pos, value
                        except EOFError:
                            break

    def iter(self):
        if self._data_id is None:
            for pos, value in itertools.chain(self._iter_file(), self._list):
                if not self._slice:
                    yield pos, value
                elif all(
                    [index_in_slice(s, i) for s, i in zip(self._slice, pos)]):
                    if self.inner_shape:
                        yield pos, value[self._slice[len(pos):]]
                    else:
                        yield pos, value
        else:
            server, record_id, key = self._data_id
            with ZMQContextManager(zmq.DEALER,
                                   connect=server,
                                   socket=self._sock) as socket:
                iter_id = b''
                start = 0
                try:
                    while True:
                        socket.send_pyobj({
                            'method': 'bufferlist_iter',
                            'record_id': record_id,
                            'iter_id': iter_id,
                            'key': key,
                            'slice': self._slice,
                            'start': start
                        })
                        iter_id, lst, finished = socket.recv_pyobj()
                        start += len(lst)
                        yield from lst
                        if finished:
                            break
                finally:
                    if iter_id:
                        socket.send_pyobj({
                            'method': 'bufferlist_iter_exit',
                            'iter_id': iter_id
                        })

    def value(self):
        d = []
        for pos, value in self.iter():
            d.append(value)
        return d

    def pos(self):
        p = []
        for pos, value in self.iter():
            p.append(pos)
        return p

    def items(self):
        p, d = [], []
        for pos, value in self.iter():
            p.append(pos)
            d.append(value)
        return p, d

    def toarray(self):
        pos, data = self.items()
        if self._slice:
            pos = np.asarray(pos)
            lu = tuple(np.min(pos, axis=0))
            rd = tuple(np.max(pos, axis=0) + 1)
            pos = np.asarray(pos) - np.asarray(lu)
            shape = []
            for k, (s, i, j) in enumerate(zip(self._slice, rd, lu)):
                if s.step is not None:
                    pos[:, k] = pos[:, k] / s.step
                    shape.append(round(np.ceil((i - j) / s.step)))
                else:
                    shape.append(i - j)
            shape = tuple(shape)
        else:
            shape = tuple([i - j for i, j in zip(self.rd, self.lu)])
            pos = np.asarray(pos) - np.asarray(self.lu)
        data = np.asarray(data)
        inner_shape = data.shape[1:]
        x = np.full(shape + inner_shape, np.nan, dtype=data[0].dtype)
        x.__setitem__(tuple(pos.T), data)
        return x

    def _full_slice(self, slice_tuple: slice
                    | tuple[slice | int | EllipsisType, ...]):
        ndim = len(self.lu)
        if self.inner_shape:
            ndim += len(self.inner_shape)

        if isinstance(slice_tuple, slice):
            slice_tuple = (
                slice_tuple, ) + (slice(0, sys.maxsize, 1), ) * (ndim - 1)
        if slice_tuple is Ellipsis:
            slice_tuple = (slice(0, sys.maxsize, 1), ) * ndim
        else:
            head, tail = (), ()
            for i, s in enumerate(slice_tuple):
                if s is Ellipsis:
                    head = slice_tuple[:i]
                    tail = slice_tuple[i + 1:]
                    break
            else:
                head = slice_tuple
                tail = ()
            slice_tuple = head + (slice(
                0, sys.maxsize, 1), ) * (ndim - len(head) - len(tail)) + tail
        slice_list = []
        contract = []
        reversed = []
        for i, s in enumerate(slice_tuple):
            if isinstance(s, int):
                if s >= 0:
                    slice_list.append(slice(s, s + 1, 1))
                elif i < len(self.lu):
                    s = self.rd[i] + s
                    slice_list.append(slice(s, s + 1, 1))
                else:
                    slice_list.append(slice(s, s - 1, -1))
                contract.append(i)
            else:
                start, stop, step = s.start, s.stop, s.step
                if step is None:
                    step = 1
                if step < 0 and i < len(self.lu):
                    step = -step
                    reversed.append(i)
                    if start is None and stop is None:
                        start, stop = 0, sys.maxsize
                    elif start is None:
                        start, stop = self.lu[i], sys.maxsize
                    elif stop is None:
                        start, stop = 0, start + self.lu[i]
                    else:
                        start, stop = stop + self.lu[i] + 1, start + self.lu[
                            i] + 1

                if start is None:
                    start = 0
                elif start < 0 and i < len(self.lu):
                    start = self.rd[i] + start
                if step is None:
                    step = 1
                if stop is None:
                    stop = sys.maxsize
                elif stop < 0 and i < len(self.lu):
                    stop = self.rd[i] + stop

                slice_list.append(slice(start, stop, step))
        return tuple(slice_list), contract, reversed

    def __getitem__(self, slice_tuple: slice | EllipsisType
                    | tuple[slice | int | EllipsisType, ...]):
        self._slice, contract, reversed = self._full_slice(slice_tuple)
        ret = self.toarray()
        slices = []
        for i, s in enumerate(self._slice):
            if i in contract:
                slices.append(0)
            elif isinstance(s, slice):
                if i in reversed:
                    slices.append(slice(None, None, -1))
                else:
                    slices.append(slice(None, None, 1))
        ret = ret.__getitem__(tuple(slices))
        self._slice = None
        return ret


class Record():

    def __init__(self, id, database, description=None):
        self.id = id
        self.database = database
        self.description = description
        self._items = {}
        self._pos = []
        self._last_vars = set()
        self._file = None
        self._sock = None

        for name, value in self.description['intrinsic_loops'].items():
            self._items[name] = value

        for level, range_list in description['loops'].items():
            for name, iterable in range_list:
                if name in self.description['independent_variables']:
                    self._items[name] = iterable

        if self.is_local_record():
            self.database = Path(self.database)
            self._file = random_path(self.database / 'objects')
            self._file.parent.mkdir(parents=True, exist_ok=True)

    def __getstate__(self) -> dict:
        return {
            'id': self.id,
            'description': self.description,
            '_items': self._items,
        }

    def __setstate__(self, state: dict):
        self.id = state['id']
        self.description = state['description']
        self._items = state['_items']
        self._pos = []
        self._last_vars = set()
        self.database = None
        self._file = None
        self._sock = None

    @property
    def axis(self):
        return self.description.get('axis', {})

    def config(self, cls=dict, session=None, datapath=None):
        config_id = self.description.get('config', None)
        if config_id is None:
            return None
        if isinstance(config_id, dict):
            return cls(config_id)
        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database,
                                   socket=self._sock) as socket:
                socket.send_pyobj({
                    'method': 'config_get',
                    'config_id': config_id
                })
                buf = socket.recv_pyobj()
                return cls(pickle.loads(lzma.decompress(buf)))
        else:
            assert session is not None, "session is required for local record"
            assert datapath is not None, "datapath is required for local record"
            config = get_config(session, config_id, base=datapath / 'objects')
            session.commit()
            return cls(config)

    def scripts(self, session=None):
        scripts = self.description['entry']['scripts']
        if isinstance(scripts, list):
            return scripts
        else:
            cell_id = scripts

        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database,
                                   socket=self._sock) as socket:
                socket.send_pyobj({
                    'method': 'notebook_history',
                    'cell_id': cell_id
                })
                return socket.recv_pyobj()
        elif self.is_local_record():
            from .models import Cell
            assert session is not None, "session is required for local record"
            cell = session.get(Cell, cell_id)
            return [
                cell.input.text
                for cell in cell.notebook.cells[1:cell.index + 2]
            ]

    def is_local_record(self):
        return not self.is_cache_record() and not self.is_remote_record()

    def is_cache_record(self):
        return self.database is None

    def is_remote_record(self):
        return isinstance(self.database,
                          str) and self.database.startswith("tcp://")

    def __del__(self):
        self.flush()

    def __getitem__(self, key):
        return self.get(key, buffer_to_array=True)

    def get(self, key, default=_not_given, buffer_to_array=False):
        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database,
                                   socket=self._sock) as socket:
                socket.send_pyobj({
                    'method': 'record_getitem',
                    'record_id': self.id,
                    'key': key
                })
                ret = socket.recv_pyobj()
                if isinstance(ret, BufferList):
                    ret._data_id = self.database, self.id, key
                    ret._sock = socket
                    if buffer_to_array:
                        return ret.toarray()
                    else:
                        return ret
                elif isinstance(ret, Space) and buffer_to_array:
                    return ret.toarray()
                else:
                    return ret
        else:
            if default is _not_given:
                d = self._items.get(key)
            else:
                d = self._items.get(key, default)
            if isinstance(d, BufferList):
                if isinstance(d.file, str):
                    d.file = self._file.parent.parent.parent.parent / d.file
                if buffer_to_array:
                    return d.toarray()
                else:
                    return d
            elif isinstance(d, Space):
                if buffer_to_array:
                    return d.toarray()
                else:
                    return d
            else:
                return d

    def keys(self):
        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database,
                                   socket=self._sock) as socket:
                socket.send_pyobj({
                    'method': 'record_keys',
                    'record_id': self.id
                })
                return socket.recv_pyobj()
        else:
            return list(self._items.keys())

    def append(self, level, step, position, variables):
        if level < 0:
            self.flush()
            return

        for key in set(variables.keys()) - self._last_vars:
            if key not in self.axis:
                self.axis[key] = tuple(range(level + 1))

        self._last_vars = set(variables.keys())

        if level >= len(self._pos):
            l = level + 1 - len(self._pos)
            self._pos.extend(([0] * (l - 1)) + [position])
            pos = tuple(self._pos)
        elif level == len(self._pos) - 1:
            self._pos[-1] = position
            pos = tuple(self._pos)
        else:
            self._pos = self._pos[:level + 1]
            self._pos[-1] = position
            pos = tuple(self._pos)
            self._pos[-1] += 1

        for key, value in variables.items():
            if self.axis[key] == ():
                if key not in self._items:
                    self._items[key] = value
            elif level == self.axis[key][-1]:
                if key not in self._items:
                    if self.is_local_record():
                        bufferlist_file = random_path(self.database /
                                                      'objects')
                        bufferlist_file.parent.mkdir(parents=True,
                                                     exist_ok=True)
                        self._items[key] = BufferList(bufferlist_file)
                    else:
                        self._items[key] = BufferList()
                    self._items[key].lu = pos
                    self._items[key].rd = tuple([i + 1 for i in pos])
                    self._items[key].append(pos, value, self.axis[key])
                elif isinstance(self._items[key], BufferList):
                    self._items[key].append(pos, value, self.axis[key])

    def flush(self):
        if self.is_remote_record() or self.is_cache_record():
            return

        for key, value in self._items.items():
            if isinstance(value, BufferList):
                value.flush()

        with open(self._file, 'wb') as f:
            dill.dump(self, f)

    def delete(self):
        if self.is_remote_record():
            with ZMQContextManager(zmq.DEALER,
                                   connect=self.database,
                                   socket=self._sock) as socket:
                socket.send_pyobj({
                    'method': 'record_delete',
                    'record_id': self.id
                })
        elif self.is_local_record():
            for key, value in self._items.items():
                if isinstance(value, BufferList):
                    value.delete()
            self._file.unlink()

    def export(self, file):
        with zipfile.ZipFile(file,
                             'w',
                             compression=zipfile.ZIP_DEFLATED,
                             compresslevel=9) as z:
            items = {}
            for key in self.keys():
                value = self.get(key)
                if isinstance(value, BufferList):
                    v = BufferList()
                    v.lu = value.lu
                    v.rd = value.rd
                    v.inner_shape = value.inner_shape
                    items[key] = v
                    with z.open(f'{key}.buf', 'w') as f:
                        for pos, data in value.iter():
                            dill.dump((pos, data), f)
                else:
                    items[key] = value
            with z.open('record.pkl', 'w') as f:
                self.description['entry']['scripts'] = self.scripts()
                dill.dump((self.description, items), f)
            with z.open('config.pkl', 'w') as f:
                f.write(dill.dumps(self.config()))

    @classmethod
    def load(cls, file: str):
        with zipfile.ZipFile(file, 'r') as z:
            with z.open('record.pkl', 'r') as f:
                description, items = dill.load(f)
            with z.open('config.pkl', 'r') as f:
                config = dill.load(f)
            description['config'] = config
            record = cls(None, None, description)
            for key, value in items.items():
                if isinstance(value, BufferList):
                    value.file = file, f'{key}.buf'
                record._items[key] = value
        return record

    def __repr__(self):
        if self.is_remote_record() and not self._items:
            self._items = {key: None for key in self.keys()}
        return f"<Record: id={self.id} app={self.description['app']}, keys={self._items.keys()}>"

    # def _repr_html_(self):
    #     return f"""
    #     <h3>Record: id={self.id}, app={self.description['app']}</h3>
    #     <p>keys={self.keys()}</p>
    #     <p>axis={self.axis}</p>
    #     """



from sqlalchemy import Engine
from sqlalchemy.orm import Session

from .base import Base
from .file import File, FileChunk
from .ipy import Cell, InputText, Notebook
from .record import Record
from .tag import Tag, has_tags


def create_tables(url: str):
    from sqlalchemy import create_engine
    engine = create_engine(url)
    Base.metadata.create_all(engine)


def create_session(url: str) -> Session:
    from sqlalchemy import create_engine
    from sqlalchemy.orm import sessionmaker
    engine = create_engine(url)
    session = sessionmaker(bind=engine)
    return session()
